# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['ncall']

# %% ../nbs/00_core.ipynb 4
import inspect
from typing import Optional, Callable
from atyp import (
    AnyQ, BoolQ, FloatQ, StrLikeQ,    
    ClsStrQ, ClsBool, ClsFloat,
)
from indoc import indoc
from nymstr import nymstr
from synset import synset

from ispec.sig import safegetsig
from ispec.dun import getname
from idfunc import passable, identity

# %% ../nbs/00_core.ipynb 6
@indoc
class ncall(synset):
    '''ncall (Named Callable Class) extends the nymstr type to handle aliased callables.

    Parameters
    ----------
    canon : str
        The canonical name of the string.

    func : Callable
        The callable to be aliased.

    *synonyms : Iterable[str]
        The synonyms of the string. They will be added to the set alongside the canonical name.

    icase : bool, default: False
        Whether to ignore case when matching, by default False.

    ispace : bool, default: False
        Whether to ignore spaces when matching, by default False.

    iunder : bool, default: False
        Whether to ignore underscores when matching, by default False.
        
    cutoff : float, default: 1
        The cutoff for fuzzy matching, by default 1.

    itemattr : ClsStrQ, default: 'synonyms'
        The attribute which holds the item being named, by default 'synonyms'.

    synsattr: ClsStrQ, default: 'synonyms'
        The attribute which holds the synonyms set, by default 'synonyms'.

    fullrepr : ClsBool, default:  False
        Whether or not to show the full representation, by default False.

    funcrepr : ClsBool, default: False
        Whether or not to show the functoin name in the representation, by default False.

    fuzzsyns : ClsBool, default: True
        Whether or not to make all synonyms fuzzy, by default True.

    Attributes
    ----------
    func : Callable
        The callable to be aliased.

    fname : str
        The name of the callable.

    fsig : inspect.Signature
        The signature of the callable.

    sigstr : str
        The signature of the callable as a string.

    syns : set
        The dynamic `synsattr` attribute representing the synonym set.

    alts : set
        The synonyms (canon excluded).
        
    nalts : int
        The number of synonyms (canon excluded).

    raw : str
        The raw string.

    canon : str
        The canonical string (i.e. the raw string))
    
    known : set
        The set of known strings (canon and synonyms included)

    snake : str
        The snake-cased string.

    item : set
        The item which was named. In this case it is a set of synonyms (canon included).

    aspec : list | tuple | dict, default: ('icase', 'ispace', 'iunder', 'cutoff', 'itemattr', 'synsattr', 'fullrepr', 'fuzzsyns', 'funcrepr', )
        The attribute specification. Must be either an iterable of strings 
        or a dictionary of `str: Any`. Defaults to an empty tuple.
    
    dspec : list | tuple | dict, default: ('itemattr', 'synsattr',)
        The dynamic attribute specification. Must be either an iterable of strings 
        or a dictionary of `str: str`. Defaults to an empty tuple.

    specs: {('aspec', ), ('aspec', 'dspec', ), ('aspec', 'dspec), (), }, default: (Spec.aspec.value, Spec.dspec.value)
        The attribute specifications to use.
        
    __readonly__ : tuple, default: (Spec.aspec.value, Spec.dspec.value)
        Attributes that cannot be set.

    Methods
    -------
    __call__(*args, **kwargs):
        Call the instance's item.

    iscanon(other: str) -> bool:
        Check if `other` is the instance's canonical name.
        
    isknown(other: str, extra: SetQ = None) -> bool:
        Check if the fuzzified `other` matches any of the known (and extra) strings.

    issyn(other: str, extra: Optional[set] = None) -> bool:
        Check if other is a synonym of the instance.

    issyns(other: str) -> bool:
        Check if other's synonym set is equivalent.

    obj2set(other) -> set:
        Convert `other` to a set.

    isitem(other: AnyQ) -> bool:
        Check if `other` is the instance's item.

    tostr() -> str:
        Convert the instance to a string.

    prep(s: StrLikeQ = None) -> str:
        Prepare a string by applying case and space insensitivity rules.

    find(strseq: Iterable[StrLike], **kwargs) -> List[str]:
        Find the closest matches (same-ish) in the provided sequence using fuzzy matching.

    iseq(other: str) -> bool:
        Check if the provided string is an alias of this entity using fuzzy matching.

    __eq__(other):
        Check if the provided value is an alias of this entity using fuzzy matching.

    __str__():
        Return the raw string.

    __hash__():
        Return the hash of the raw string.

    __setattr__(name, value):
        Set the class attribute if it's not read-only.

    getattrkeys(spec: str, dyn: bool = False) -> tuple[str, ...]:
        Return attribute keys stored in `spec`.

    getattrvals(self, spec: str, dyn: bool = False) -> tuple[Any, ...]:
        Return default attribute values stored in `spec`.
    
    skeys() -> tuple[str, ...]:
        Return all attribute keys stored for each spec stored in `specs`.

    svals() -> tuple[str, ...]:
        Return default attribute values stored for each spec stored in `specs`.
    
    getattrs(**kwargs):
        Get instance parameters with optional overrides.
        
    makesame(*args, **kwargs):
        Call class constructor with the same attributes as the current instance.
    
    isinst(other):
        Check if the provided value is an instance of this class.
        
    sameattrs(other):
        Check if the provided value is an instance of this class with the same attributes.
        
    diffattrs(other):
        Check if the provided value is an instance of this class with different attributes.

    getattrname(dattr: str) -> str:
        Get the name of the dynamic attribute.
    
    getdattr(dattr: str, default: Any = None) -> Any:
        Get the value of the dynamic attribute.

    setdattr(dattr: str, value: Any = None):
        Set the value of the dynamic attribute.

    update_attrname(dattr: str, **kwargs):
        Update the name of the dynamic attribute.
    
    update_dattrval(aname: str, **kwargs):
        Update the value of the dynamic attribute.
    
    update_dattr(dattr: str, **kwargs):
        Update the name and value of the dynamic attribute.

    update_aspec(**kwargs):
        Update the attribute specification.

    update_dspec(**kwargs):
        Update the dynamic attribute specification.

    update_specs(**kwargs):
        Update the specifications in `specs` e.g. `aspec` and / or `dspec`.

    getclsattr(attr: str, default: Any = None) -> Any:
        Get the value of the class attribute.

    setclsattr(attr: str, val: Any = None):
        Set the value of the class attribute.

    copy():
        Return a shallow copy of the instance.
    
    deepcopy():
        Return a deep copy of the instance.

    __contains__(syn: str):
        Check if a synonym is in the set.
    
    __len__():
        Return the number of synonyms in the set (canon included).
    
    __str__():
        Return the string for the canonical name.

    __iter__():
        Return an iterator over the set.
    
    __repr__():
        Return a string representation of the object.

    __and__(other):
        Return the intersection of two sets.

    __ge__(other):
        Check if one set is a superset of another.
        
    __gt__(other):
        Check if one set is a proper superset of another.
        
    __iand__(other):
        Update a set with the intersection of itself and another.
        
    __ior__(other):
        Update a set with the union of itself and another.
        
    __isub__(other):
        Remove all elements of another set from this set.
        
    __ixor__(other):
        Update a set with the symmetric difference of itself and another.
        
    __le__(other):
        Check if one set is a subset of another.
        
    __lt__(other):
        Check if one set is a proper subset of another.
        
    __or__(other):
        Return the union of sets as a new set.
        
    __rand__(other):
        Return the intersection of two sets as a new set.
        
    __ror__(other):
        Return the union of sets as a new set.

    __xor__(other):
        Return the symmetric difference of two sets as a new set.

    __sub__(other):
        Return the difference of two sets as a new set.
        
    __rsub__(other):
        Return the difference of two or more sets as a new set.
        
    __rxor__(other):
        Return the symmetric difference of two sets as a new set.
        
    add(syn: str):
        Add a synonym to the set.
        
    remove(syn: str):
        Remove a synonym from the set.
        
    clear():
        Remove all synonyms from the set.
        
    difference(*others):
        Return the difference of two or more sets as a new set.
        
    difference_update(*others):
        Remove all elements of another set from this set.
        
    discard(syn: str):
        Remove a synonym from the set.
        
    intersection(*others):
        Return the intersection of two or more sets as a new set.
        
    intersection_update(*others):
        Update a set with the intersection of itself and another.
        
    isdisjoint(other):
        Return True if two sets have a null intersection.
        
    issubset(other):
        Report whether another set contains this set.
        
    issuperset(other):
        Report whether this set contains another set.
        
    pop():
        Remove and return an arbitrary set element.
        
    remove(syn: str):
        Remove a synonym from the set.
        
    symmetric_difference(other):
        Return the symmetric difference of two sets as a new set.
        
    symmetric_difference_update(other):
        Update a set with the symmetric difference of itself and another.
        
    union(*others):
        Return the union of sets as a new set.
        
    update(*others):
        Update a set with the union of itself and others.
        
    ''' 
    icase:  ClsBool  = True # Whether to ignore case when matching
    ispace: ClsBool  = True # Whether to ignore spaces when matching
    iunder: ClsBool  = True # Whether to ignore underscores when matching
    cutoff: ClsFloat = 0.8  # The cutoff for fuzzy matching
    
    itemattr: ClsStrQ = 'callable'
    synsattr: ClsStrQ = 'synonyms'
    fullrepr: ClsBool = False
    fuzzsyns: ClsBool  = True # Make all synonyms fuzstrs

    aspec = ('icase', 'ispace', 'iunder', 'cutoff', 'itemattr', 'synsattr', 'fullrepr', 'fuzzsyns', 'funcrepr', )
    dspec = ('itemattr', 'synsattr',)

    funcrepr: ClsBool = False

    @property
    def func(self) -> str:
        '''The dynamic `itemattr` i.e. `getattr(self, self.itemattr, None)`'''
        return self.getdattr('itemattr', None)
    
    def __new__(
        cls, canon: str, func: Callable, *synonyms, 
        icase:  BoolQ  = True,           # Case insensitive
        ispace: BoolQ  = True,           # Space insensitive
        iunder: BoolQ  = True,           # Underscore insensitive
        cutoff: FloatQ = 0.8,            # Fuzzy cutoff
        itemattr: StrLikeQ = 'callable', # Dynamic item attribute
        synsattr: StrLikeQ = 'synonyms', # Dynamic synonym attribute
        fullrepr: BoolQ = True,          # Full representation
        fuzzsyns: BoolQ  = True,         # Make all synonyms fuzstrs
        funcrepr: BoolQ = False,         # Show function name in repr
        **kwargs
    ):
        # Put all variables in a dictionary for easy access
        fnset = {getname(func), } or set()
        synonyms = ({canon, *synonyms} if synonyms else {canon, }) | fnset
        kwargs.update(
            icase=icase, ispace=ispace, iunder=iunder, cutoff=cutoff,
            itemattr=itemattr, synsattr=synsattr, fullrepr=fullrepr,
            fuzzsyns=fuzzsyns, funcrepr=funcrepr,
        )
        kwargs[itemattr] = func
        kwargs[synsattr] = synonyms
        obj = super().__new__(cls, canon, *synonyms, **kwargs)
        return obj
    
    def __init__(
        self, canon: str, func: Callable, *synonyms, 
        icase:  BoolQ   = True,          # Case insensitive
        ispace: BoolQ   = True,          # Space insensitive
        iunder: BoolQ   = True,          # Underscore insensitive
        cutoff: FloatQ  = 0.8,           # Fuzzy cutoff
        itemattr: StrLikeQ = 'callable', # Dynamic item attribute
        synsattr: StrLikeQ = 'synonyms', # Dynamic synonym attribute
        fullrepr: BoolQ = True,          # Full representation
        fuzzsyns: BoolQ  = True,         # Make all synonyms fuzstrs
        funcrepr: BoolQ = False,         # Show function name in repr
        **kwargs
    ):
        # Put all variables in a dictionary for easy access        
        fnset = {getname(func), } or set()
        synonyms = ({canon, *synonyms} if synonyms else {canon, }) | fnset
        kwargs.update(
            icase=icase, ispace=ispace, iunder=iunder, cutoff=cutoff,
            itemattr=itemattr, synsattr=synsattr, fullrepr=fullrepr,
            fuzzsyns=fuzzsyns, funcrepr=funcrepr,
        )
        kwargs[itemattr] = func
        kwargs[synsattr] = synonyms
        super().__init__(canon, *synonyms, **kwargs)

    @property
    def fname(self):
        return getname(self.func)
    
    @property
    def fsig(self) -> inspect.Signature:
        '''Return the signature function'''
        return safegetsig(self.func)
    
    @property
    def sigstr(self) -> str:
        '''Return the function signature as a string'''
        return str(self.fsig)
    
    @property
    def known(self) -> set: 
        return {self.canon, *self.alts, *(self.fname or set({}))}
    
    def isitem(self, other: AnyQ) -> bool:
        '''Check if `other` is the instance's item'''        
        if isinstance(other, type(self)) or hasattr(other, 'func'):
            return self.func == other.func
        return self.func == other

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)
    
    def __repr__(self):
        '''Return a string representation of the object.'''        
        if not self.fullrepr:
            return f'{self.canon}(*args, **kwargs)'
        if self.funcrepr:
            return f''''{self.canon}': {self.fname}{self.sigstr}'''
        return f'{self.canon}{self.sigstr}'
